/* 
   sane-troubleshoot: Find out what's wrong with your SANE installation
   Copyright (C) 2003 Henning Meier-Geinitz <henning@meier-geinitz.de>

   This part of the program reads the backend descriptions. It was taken
   from sane-desc.c (sane-backends), Copyright (C) 2002, 2003 Henning
   Meier-Geinitz <henning@meier-geinitz.de>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.
*/

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <limits.h>
#include <ctype.h>
#include <time.h>
#include <gtk/gtk.h>
#include "sane.h"

#include "descriptions.h"

#ifndef PATH_MAX
# define PATH_MAX 1024
#endif

#define DBG_ERR current_debug_level = 0; debug_call
#define DBG_WARN current_debug_level = 1; debug_call
#define DBG_INFO current_debug_level = 2; debug_call
#define DBG_DBG current_debug_level = 3; debug_call

static char *program_name;
static int debug = 0;
static int current_debug_level = 0;
static char *search_dir = 0;
static backend_entry *first_backend = 0;

const char *
skip_whitespace (const char *str)
{
  while (str && *str && isspace (*str))
    ++str;
  return str;
}

const char *
get_string (const char *str, char **string_const)
{
  const char *start;
  size_t len;

  str = skip_whitespace (str);

  if (*str == '"')
    {
      start = ++str;
      while (*str && *str != '"')
	++str;
      len = str - start;
      if (*str == '"')
	++str;
      else
	start = 0;		/* final double quote is missing */
    }
  else
    {
      start = str;
      while (*str && !isspace (*str))
	++str;
      len = str - start;
    }
  if (start)
    *string_const = strndup (start, len);
  else
    string_const = 0;
  return str;
}

char *
read_line (char *str, int n, FILE *stream)
{
   char* rc;
   char* start;
   int   len;

      /* read line from stream */
   rc = fgets( str, n, stream);
   if (rc == NULL)
      return NULL;

      /* remove ending whitespaces */
   len = strlen( str);
   while( (0 < len) && (isspace( str[--len])) )
      str[len] = '\0';

      /* remove starting whitespaces */
   start = str;
   while( isspace( *start))
      start++;

   if (start != str) 
      do {
         *str++ = *start++;
      } while( *str);

   return rc;
}



static void
debug_call (const char *fmt, ...)
{
  va_list ap;
  char *level_txt;

  va_start (ap, fmt);
  if (debug >= current_debug_level)
    {
      /* print to stderr */
      switch (current_debug_level)
	{
	case 0:
	  level_txt = "ERROR:";
	  break;
	case 1:
	  level_txt = "Warning:";
	  break;
	case 2:
	  level_txt = "Info:";
	  break;
	default:
	  level_txt = "";
	  break;
	}
      fprintf (stderr, "[%s] %8s ", program_name, level_txt);
      vfprintf (stderr, fmt, ap);
    }
  va_end (ap);
}

static int
char_compare (char char1, char char2)
{
  char1 = toupper (char1);
  char2 = toupper (char2);

  if (char1 < char2)
    return -1;
  else if (char1 > char2)
    return 1;
  else
    return 0;
}

static int
num_compare (char *num_string1, char *num_string2)
{
  int num1 = atoi (num_string1);
  int num2 = atoi (num_string2);
  if (num1 < num2)
    return -1;
  else if (num1 > num2)
    return 1;
  else
    return 0;
}

/* Compare two strings, try to sort numbers correctly (600 < 1200) */
static int
string_compare (char *string1, char *string2)
{
  int count = 0;
  int compare = 0;
  while (string1[count] && string2[count])
    {
      if (isdigit (string1[count]) && isdigit (string2[count]))
	compare = num_compare (&string1[count], &string2[count]);
      else
	compare = char_compare (string1[count], string2[count]);
      if (compare != 0)
	return compare;
      count++;
    }
  return char_compare (string1[count], string2[count]);
}

/* Add URLs to the end of the list if they are unique */
static url_entry *
update_url_list (url_entry * first_url, char *new_url)
{
  url_entry *url = first_url;
  SANE_Bool found = SANE_FALSE;

  while (url && url->name)
    {
      if (string_compare (url->name, new_url) == 0)
	found = SANE_TRUE;
      url = url->next;
    }
  if (found)
    return first_url;

  url = first_url;
  if (url)
    {
      while (url->next)
	url = url->next;
      url->next = calloc (1, sizeof (url_entry));
      url = url->next;
    }
  else
    {
      first_url = calloc (1, sizeof (url_entry));
      url = first_url;
    }
  if (!url)
    {
      DBG_ERR ("update_url_list: couldn't calloc url_entry\n");
      exit (1);
    }
  url->name = new_url;
  return first_url;
}

/* Get the next token, ignoring escaped quotation marks */
static const char *
get_token (const char *str, char **string_const)
{
  const char *start;
  size_t len;

  str = skip_whitespace (str);

  if (*str == '"')
    {
      start = ++str;
      while (*str && (*str != '"' || *(str - 1) == '\\'))
	++str;
      len = str - start;
      if (*str == '"')
	++str;
      else
	start = 0;		/* final double quote is missing */
    }
  else
    {
      start = str;
      while (*str && !isspace (*str))
	++str;
      len = str - start;
    }
  if (start)
    *string_const = strndup (start, len);
  else
    string_const = 0;
  return str;
}

/* Checks a line for a keyword token and determines keyword/string argument */
static SANE_Status
read_keyword (SANE_String line, SANE_String keyword_token,
	      parameter_type p_type, void *argument)
{
  SANE_String_Const cp;
  SANE_Char *word;

  word = 0;

  cp = get_token (line, &word);

  if (strcmp (word, keyword_token) != 0)
    return SANE_STATUS_INVAL;

  free (word);
  word = 0;

  switch (p_type)
    {
    case param_none:
      return SANE_STATUS_GOOD;
    case param_string:
      {
	char *pos;
	cp = get_token (cp, &word);
	/* remove escaped quotations */
	while ((pos = strstr (word, "\\\"")) != 0)
	  *pos = ' ';

	DBG_DBG ("read_keyword: set entry `%s' to `%s'\n", keyword_token,
		 word);
	*(SANE_String *) argument = strdup (word);
	break;
      }
    default:
      DBG_ERR ("read_keyword: unknown param_type %d\n", p_type);
      return SANE_STATUS_INVAL;
    }

  if (word)
    free (word);
  word = 0;
  return SANE_STATUS_GOOD;
}

/* Read and interprete the .desc files */
static SANE_Bool
read_files (void)
{
  struct stat stat_buf;
  DIR *dir;
  struct dirent *dir_entry;
  FILE *fp;
  char file_name[PATH_MAX];
  SANE_Char line[PATH_MAX], *word;
  SANE_String_Const cp;
  backend_entry *current_backend = 0;
  type_entry *current_type = 0;
  mfg_entry *current_mfg = 0;
  model_entry *current_model = 0;
  enum level current_level = level_backend;

  DBG_INFO ("looking for .desc files in `%s'\n", search_dir);
  if (stat (search_dir, &stat_buf) < 0)
    {
      DBG_ERR ("cannot stat `%s' (%s)\n", search_dir, strerror (errno));
      return SANE_FALSE;
    }
  if (!S_ISDIR (stat_buf.st_mode))
    {
      DBG_ERR ("`%s' is not a directory\n", search_dir);
      return SANE_FALSE;
    }
  if ((dir = opendir (search_dir)) == 0)
    {
      DBG_ERR ("cannot read directory `%s' (%s)\n", search_dir,
	       strerror (errno));
      return SANE_FALSE;
    }

  while ((dir_entry = readdir (dir)) != NULL)
    {
      if (strlen (dir_entry->d_name) > 5 &&
	  strcmp (dir_entry->d_name + strlen (dir_entry->d_name) - 5,
		  ".desc") == 0)
	{
	  if (strlen (search_dir)
	      + strlen (dir_entry->d_name) + 1 + 1 > PATH_MAX)
	    {
	      DBG_ERR ("filename too long\n");
	      return SANE_FALSE;
	    }
	  sprintf (file_name, "%s/%s", search_dir, dir_entry->d_name);
	  DBG_INFO ("-> reading desc file: %s\n", file_name);
	  fp = fopen (file_name, "r");
	  if (!fp)
	    {
	      DBG_ERR ("can't open desc file: %s (%s)\n", file_name,
		       strerror (errno));
	      return SANE_FALSE;
	    }
	  current_backend = 0;
	  current_type = 0;
	  current_mfg = 0;
	  current_model = 0;
	  while (read_line (line, sizeof (line), fp))
	    {
	      char *string_entry = 0;
	      word = 0;

	      cp = get_token (line, &word);
	      if (!word || cp == line)
		{
		  DBG_DBG ("ignoring empty line\n");
		  if (word)
		    free (word);
		  word = 0;
		  continue;
		}
	      if (word[0] == ';')
		{
		  DBG_DBG ("ignoring comment line\n");
		  free (word);
		  word = 0;
		  continue;
		}
	      DBG_DBG ("line: %s\n", line);

	      if (read_keyword (line, ":backend", param_string, &string_entry)
		  == SANE_STATUS_GOOD)
		{
		  backend_entry *be = first_backend, *prev_be = 0, *new_be =
		    0;
		  DBG_INFO ("creating backend entry `%s'\n", string_entry);

		  new_be = calloc (1, sizeof (backend_entry));
		  if (!new_be)
		    {
		      DBG_ERR ("calloc failed (%s)\n", strerror (errno));
		      return SANE_FALSE;
		    }
		  new_be->name = string_entry;
		  new_be->status = status_unknown;
		  new_be->new = SANE_FALSE;

		  if (!be)
		    {
		      first_backend = new_be;
		      be = new_be;
		    }
		  else
		    {
		      while (be)
			{
			  int compare =
			    string_compare (new_be->name, be->name);
			  if (compare <= 0)
			    {
			      backend_entry *be_tmp = be;
			      be = new_be;
			      be->next = be_tmp;
			      if (!prev_be)
				first_backend = be;
			      else
				prev_be->next = be;
			      break;
			    }
			  prev_be = be;
			  be = be->next;
			}
		      if (!be)	/* last entry */
			{
			  prev_be->next = new_be;
			  be = prev_be->next;
			}
		    }
		  current_backend = be;
		  current_type = 0;
		  current_mfg = 0;
		  current_model = 0;
		  current_level = level_backend;
		  continue;
		}
	      if (!current_backend)
		{
		  DBG_ERR ("use `:backend' keyword first\n");
		  return SANE_FALSE;
		}
	      if (read_keyword (line, ":version", param_string, &string_entry)
		  == SANE_STATUS_GOOD)
		{
		  if (current_backend->version)
		    {
		      DBG_WARN ("overwriting version of backend `%s' to `%s'"
				"(was: `%s')\n",
				current_backend->name, string_entry,
				current_backend->version, current_backend->version);
		    }

		  DBG_INFO ("setting version of backend `%s' to `%s'\n",
			    current_backend->name, string_entry);
		  current_backend->version = string_entry;
		  continue;
		}
	      if (read_keyword (line, ":status", param_string, &string_entry)
		  == SANE_STATUS_GOOD)
		{
		  switch (current_level)
		    {
		    case level_backend:
		      if (current_backend->status != status_unknown)
			{
			  DBG_WARN ("overwriting status of backend `%s'\n",
				    current_backend->name);
			}
		      if (strcmp (string_entry, ":new") == 0)
			{
			  DBG_WARN ("ignored `%s' status :new, use keyword "
				    "`:new :yes' instead\n",
				    current_backend->name);
			  current_backend->status = status_unknown;
			}
		      else if (strcmp (string_entry, ":alpha") == 0)
			{
			  DBG_INFO
			    ("setting status of backend `%s' to `alpha'\n",
			     current_backend->name);
			  current_backend->status = status_alpha;
			}
		      else if (strcmp (string_entry, ":beta") == 0)
			{
			  DBG_INFO
			    ("setting status of backend `%s' to `beta'\n",
			     current_backend->name);
			  current_backend->status = status_beta;
			}
		      else if (strcmp (string_entry, ":stable") == 0)
			{
			  DBG_INFO
			    ("setting status of backend `%s' to `stable'\n",
			     current_backend->name);
			  current_backend->status = status_stable;
			}
		      else
			{
			  DBG_ERR ("unknown status of backend `%s': `%s'\n",
				   current_backend->name, string_entry);
			  current_backend->status = status_unknown;
			  return SANE_FALSE;
			}
		      break;
		    case level_model:
		      if (current_model->status != status_unknown)
			{
			  DBG_WARN ("overwriting status of model `%s'\n",
				    current_model->name);
			}
		      if (strcmp (string_entry, ":alpha") == 0)
			{
			  DBG_INFO
			    ("setting status of model `%s' to `alpha'\n",
			     current_model->name);
			  current_model->status = status_alpha;
			}
		      else if (strcmp (string_entry, ":beta") == 0)
			{
			  DBG_INFO
			    ("setting status of model `%s' to `beta'\n",
			     current_model->name);
			  current_model->status = status_beta;
			}
		      else if (strcmp (string_entry, ":stable") == 0)
			{
			  DBG_INFO
			    ("setting status of model `%s' to `stable'\n",
			     current_model->name);
			  current_model->status = status_stable;
			}
		      else if (strcmp (string_entry, ":untested") == 0)
			{
			  DBG_INFO
			    ("setting status of model `%s' to `untested'\n",
			     current_model->name);
			  current_model->status = status_untested;
			}
		      else if (strcmp (string_entry, ":unsupported") == 0)
			{
			  DBG_INFO
			    ("setting status of model `%s' to `unsupported'\n",
			     current_model->name);
			  current_model->status = status_unsupported;
			}
		      else
			{
			  DBG_ERR ("unknown status of model `%s': `%s'\n",
				   current_model->name, string_entry);
			  current_model->status = status_unknown;
			  return SANE_FALSE;
			}

		      break;
		    default:
		      DBG_ERR ("level %d not implemented for :status\n",
			       current_level);
		      return SANE_FALSE;
		    }


		  continue;
		}
	      if (read_keyword (line, ":new", param_string, &string_entry)
		  == SANE_STATUS_GOOD)
		{
		  if (strcmp (string_entry, ":yes") == 0)
		    {
		      DBG_INFO ("backend %s is new in this SANE release\n",
				current_backend->name);
		      current_backend->new = SANE_TRUE;
		    }
		  else if (strcmp (string_entry, ":no") == 0)
		    {
		      DBG_INFO
			("backend %s is NOT new in this SANE release\n",
			 current_backend->name);
		      current_backend->new = SANE_FALSE;
		    }
		  else
		    {
		      DBG_ERR ("unknown :new parameter of backend `%s': "
			       "`%s'\n", current_backend->name, string_entry);
		      current_backend->new = SANE_FALSE;
		      return SANE_FALSE;
		    }
		  continue;
		}
	      if (read_keyword (line, ":manpage", param_string, &string_entry)
		  == SANE_STATUS_GOOD)
		{
		  if (current_backend->manpage)
		    {
		      DBG_WARN ("overwriting manpage of backend `%s' to `%s'"
				"(was: `%s')\n",
				current_backend->name, string_entry,
				current_backend->manpage);
		    }

		  DBG_INFO ("setting manpage of backend `%s' to `%s'\n",
			    current_backend->name, string_entry);
		  current_backend->manpage = string_entry;
		  continue;
		}
	      if (read_keyword
		  (line, ":devicetype", param_string,
		   &string_entry) == SANE_STATUS_GOOD)
		{
		  type_entry *type = 0;

		  type = current_backend->type;

		  DBG_INFO ("adding `%s' to list of device types of backend "
			    "`%s'\n", string_entry, current_backend->name);

		  if (type)
		    {
		      while (type->next)
			type = type->next;
		      type->next = calloc (1, sizeof (type_entry));
		      type = type->next;
		    }
		  else
		    {
		      current_backend->type = calloc (1, sizeof (type_entry));
		      type = current_backend->type;
		    }

		  type->type = type_unknown;
		  if (strcmp (string_entry, ":scanner") == 0)
		    {
		      DBG_INFO ("setting device type of backend `%s' to "
				"scanner\n", current_backend->name);
		      type->type = type_scanner;
		    }
		  else if (strcmp (string_entry, ":stillcam") == 0)
		    {
		      DBG_INFO ("setting device type of backend `%s' to "
				"still camera\n", current_backend->name);
		      type->type = type_stillcam;
		    }
		  else if (strcmp (string_entry, ":vidcam") == 0)
		    {
		      DBG_INFO ("setting device type of backend `%s' to "
				"video camera\n", current_backend->name);
		      type->type = type_vidcam;
		    }
		  else if (strcmp (string_entry, ":api") == 0)
		    {
		      DBG_INFO ("setting device type of backend `%s' to "
				"API\n", current_backend->name);
		      type->type = type_api;
		    }
		  else if (strcmp (string_entry, ":meta") == 0)
		    {
		      DBG_INFO ("setting device type of backend `%s' to "
				"meta\n", current_backend->name);
		      type->type = type_meta;
		    }
		  else
		    {
		      DBG_ERR ("unknown device type of backend `%s': `%s'\n",
			       current_backend->name, string_entry);
		      type->type = type_unknown;
		      return SANE_FALSE;
		    }
		  current_type = type;
		  current_mfg = 0;
		  current_model = 0;
		  continue;
		}
	      if (read_keyword (line, ":desc", param_string, &string_entry)
		  == SANE_STATUS_GOOD)
		{
		  if (!current_type)
		    {
		      DBG_ERR ("use `:devicetype' keyword first\n");
		      return SANE_FALSE;
		    }
		  if (current_type->type < type_meta)
		    {
		      DBG_ERR ("use `:desc' for `:api' and `:meta' only\n");
		      return SANE_FALSE;
		    }

		  if (current_type->desc)
		    {
		      DBG_WARN ("overwriting description of  device type of "
				"backend `%s' to `%s' (was: `%s')\n",
				current_backend->name, string_entry,
				current_type->desc);
		    }

		  DBG_INFO ("setting description of backend `%s' to `%s'\n",
			    current_backend->name, string_entry);
		  current_type->desc = calloc (1, sizeof (desc_entry));
		  if (!current_type->desc)
		    {
		      DBG_ERR ("calloc failed (%s)\n", strerror (errno));
		      return SANE_FALSE;
		    }
		  current_type->desc->desc = string_entry;
		  current_level = level_desc;
		  current_mfg = 0;
		  current_model = 0;
		  continue;
		}
	      if (read_keyword (line, ":mfg", param_string, &string_entry)
		  == SANE_STATUS_GOOD)
		{
		  mfg_entry *mfg = 0;

		  if (!current_type)
		    {
		      DBG_ERR ("use `:devicetype' keyword first\n");
		      return SANE_FALSE;
		    }
		  if (current_type->type >= type_meta)
		    {
		      DBG_ERR ("use `:mfg' for hardware devices only\n");
		      return SANE_FALSE;
		    }

		  mfg = current_type->mfg;
		  if (mfg)
		    {
		      while (mfg->next)
			mfg = mfg->next;
		      mfg->next = calloc (1, sizeof (mfg_entry));
		      mfg = mfg->next;
		    }
		  else
		    {
		      current_type->mfg = calloc (1, sizeof (mfg_entry));
		      mfg = current_type->mfg;
		    }

		  if (!mfg)
		    {
		      DBG_ERR ("calloc failed (%s)\n", strerror (errno));
		      return SANE_FALSE;
		    }
		  mfg->name = string_entry;
		  DBG_INFO ("adding mfg entry %s to backend `%s'\n",
			    string_entry, current_backend->name);
		  current_mfg = mfg;
		  current_model = 0;
		  current_level = level_mfg;
		  continue;
		}
	      if (read_keyword (line, ":model", param_string, &string_entry)
		  == SANE_STATUS_GOOD)
		{
		  model_entry *model = 0;

		  if (!current_type)
		    {
		      DBG_ERR ("use `:devicetype' keyword first\n");
		      return SANE_FALSE;
		    }
		  if (current_level != level_mfg
		      && current_level != level_model)
		    {
		      DBG_ERR ("use `:mfg' keyword first\n");
		      return SANE_FALSE;
		    }
		  model = current_mfg->model;
		  if (model)
		    {
		      while (model->next)
			model = model->next;
		      model->next = calloc (1, sizeof (model_entry));
		      model = model->next;
		    }
		  else
		    {
		      current_mfg->model = calloc (1, sizeof (model_entry));
		      model = current_mfg->model;
		    }

		  if (!model)
		    {
		      DBG_ERR ("calloc failed (%s)\n", strerror (errno));
		      return SANE_FALSE;
		    }
		  model->name = string_entry;
		  model->status = status_unknown;
		  DBG_INFO ("adding model entry %s to manufacturer `%s'\n",
			    string_entry, current_mfg->name);
		  current_model = model;
		  current_level = level_model;
		  continue;
		}
	      if (read_keyword
		  (line, ":interface", param_string,
		   &string_entry) == SANE_STATUS_GOOD)
		{
		  if (!current_model)
		    {
		      DBG_WARN ("ignored `%s' :interface, only allowed for "
				"hardware devices\n", current_backend->name);
		      continue;
		    }

		  if (current_model->interface)
		    {
		      DBG_WARN ("overwriting interface of model "
				"`%s' to `%s' (was: `%s')\n",
				current_model->name, string_entry,
				current_model->interface);
		    }

		  DBG_INFO ("setting interface of model `%s' to `%s'\n",
			    current_model->name, string_entry);
		  current_model->interface = string_entry;
		  continue;
		}
	      if (read_keyword (line, ":url", param_string, &string_entry)
		  == SANE_STATUS_GOOD)
		{
		  switch (current_level)
		    {
		    case level_backend:
		      current_backend->url =
			update_url_list (current_backend->url, string_entry);
		      DBG_INFO ("adding `%s' to list of urls of backend "
				"`%s'\n", string_entry,
				current_backend->name);
		      break;
		    case level_mfg:
		      current_mfg->url =
			update_url_list (current_mfg->url, string_entry);
		      DBG_INFO ("adding `%s' to list of urls of mfg "
				"`%s'\n", string_entry, current_mfg->name);
		      break;
		    case level_desc:
		      current_type->desc->url =
			update_url_list (current_type->desc->url,
					 string_entry);
		      DBG_INFO ("adding `%s' to list of urls of description "
				"for backend `%s'\n", string_entry,
				current_backend->name);
		      break;
		    case level_model:
		      current_model->url =
			update_url_list (current_model->url, string_entry);
		      DBG_INFO ("adding `%s' to list of urls of model "
				"`%s'\n", string_entry, current_model->name);
		      break;
		    default:
		      DBG_ERR ("level %d not implemented for :url\n",
			       current_level);
		      return SANE_FALSE;
		    }
		  continue;
		}
	      if (read_keyword (line, ":comment", param_string, &string_entry)
		  == SANE_STATUS_GOOD)
		{
		  switch (current_level)
		    {
		    case level_backend:
		      current_backend->comment = string_entry;
		      DBG_INFO ("setting comment of backend %s to `%s'\n",
				current_backend->name, string_entry);
		      break;
		    case level_mfg:
		      current_mfg->comment = string_entry;
		      DBG_INFO
			("setting comment of manufacturer %s to `%s'\n",
			 current_mfg->name, string_entry);
		      break;
		    case level_desc:
		      current_type->desc->comment = string_entry;
		      DBG_INFO ("setting comment of description for "
				"backend %s to `%s'\n", current_backend->name,
				string_entry);
		      break;
		    case level_model:
		      current_model->comment = string_entry;
		      DBG_INFO ("setting comment of model %s to `%s'\n",
				current_model->name, string_entry);
		      break;
		    default:
		      DBG_ERR ("level %d not implemented for `:comment'\n",
			       current_level);
		      return SANE_FALSE;
		    }
		  continue;
		}
	      DBG_ERR ("unknown keyword token in line `%s'\n", line);
	      return SANE_FALSE;
	    }			/* while (read_line) */
	  fclose (fp);
	}			/* if (strlen) */
    }				/* while (direntry) */
  if (!first_backend)
    {
      DBG_ERR ("Couldn't find any .desc file\n");
      return SANE_FALSE;
    }
  return SANE_TRUE;
}

/* Create a model_record_entry based on a model_entry */
static model_record_entry *
create_model_record (model_entry * model)
{
  model_record_entry *model_record;

  model_record = calloc (1, sizeof (model_record_entry));
  if (!model_record)
    {
      DBG_ERR ("create_model_record: couldn't calloc model_record_entry\n");
      exit (1);
    }
  model_record->name = model->name;
  model_record->status = model->status;
  model_record->interface = model->interface;
  model_record->url = model->url;
  model_record->comment = model->comment;
  return model_record;
}

/* Calculate the priority of statuses: */
/* alpha, beta, stable -> 2, untested -> 1, unsupported -> 0 */
static int
calc_priority (status_entry status)
{
  switch (status)
    {
    case status_untested:
      return 1;
    case status_unsupported:
      return 0;
    default:
      return 2;
    }
}

/* Insert model into list at the alphabetically correct position */
static model_record_entry *
update_model_record_list (model_record_entry * first_model_record,
			  model_entry * model, backend_entry * be)
{
  model_record_entry *model_record = first_model_record;

  if (!first_model_record)
    {
      /* First model for this manufacturer */
      first_model_record = create_model_record (model);
      model_record = first_model_record;
    }
  else
    {
      model_record_entry *prev_model_record = 0;

      while (model_record)
	{
	  int compare = string_compare (model->name, model_record->name);
	  if (compare <= 0)
	    {
	      model_record_entry *tmp_model_record = model_record;
	      if (compare == 0 &&
		  string_compare (model->interface, model_record->interface)
		  == 0)
		{
		  /* Two entries for the same model */
		  int new_priority = calc_priority (model->status);
		  int old_priority = calc_priority (model_record->status);
		  if (new_priority < old_priority)
		    {
		        DBG_DBG ("update_model_record_list: model %s ignored, backend %s has "
				 "higher priority\n", model->name, model_record->be->name);
			return first_model_record;
		    }
		  if (new_priority > old_priority)
		    {
		      DBG_DBG ("update_model_record_list: model %s overrides the one from backend %s\n",
			       model->name, model_record->be->name);
		      tmp_model_record = model_record->next;
		    }
		}
	      /* correct position */
	      model_record = create_model_record (model);
	      model_record->next = tmp_model_record;
	      if (!prev_model_record)
		first_model_record = model_record;
	      else
		prev_model_record->next = model_record;
	      break;
	    }
	  prev_model_record = model_record;
	  model_record = model_record->next;
	}
      if (!model_record)	/* last entry */
	{
	  prev_model_record->next = create_model_record (model);
	  model_record = prev_model_record->next;
	}
    }				/* if (first_model_record) */
  model_record->be = be;
  DBG_DBG ("update_model_record_list: added model %s\n", model->name);
  return first_model_record;
}


/* Insert manufacturer into list at the alphabetically correct position, */
/* create new record if neccessary */
static mfg_record_entry *
update_mfg_record_list (mfg_record_entry * first_mfg_record, mfg_entry * mfg,
			backend_entry * be)
{
  model_entry *model = mfg->model;
  mfg_record_entry *mfg_record = first_mfg_record;

  while (mfg_record)
    {
      if (string_compare (mfg_record->name, mfg->name) == 0)
	{
	  /* Manufacturer already exists */
	  url_entry *mfg_url = mfg->url;

	  /* Manufacturer comments and (additional) URLs? */
	  if (!mfg_record->comment)
	    mfg_record->comment = mfg->comment;
	  while (mfg_url && mfg_url->name)
	    {
	      mfg_record->url = update_url_list (mfg_record->url,
						 mfg_url->name);
	      mfg_url = mfg_url->next;
	    }
	  break;
	}
      mfg_record = mfg_record->next;
    }

  if (!mfg_record)
    {
      /* Manufacturer doesn't exist yet */
      url_entry *url = mfg->url;

      mfg_record = calloc (1, sizeof (mfg_record_entry));
      if (!mfg_record)
	{
	  DBG_ERR ("update_mfg_record_list: couldn't calloc "
		   "mfg_record_entry\n");
	  exit (1);
	}
      mfg_record->name = mfg->name;
      mfg_record->comment = mfg->comment;
      while (url)
	{
	  mfg_record->url = update_url_list (mfg_record->url, url->name);
	  url = url->next;
	}
      if (first_mfg_record != 0)
	{
	  /* We already have one manufacturer in the list */
	  mfg_record_entry *new_mfg_record = mfg_record;
	  mfg_record_entry *prev_mfg_record = 0;

	  mfg_record = first_mfg_record;

	  while (mfg_record)
	    {
	      int compare =
		string_compare (new_mfg_record->name, mfg_record->name);
	      if (compare <= 0)
		{
		  mfg_record_entry *tmp_mfg_record = mfg_record;
		  mfg_record = new_mfg_record;
		  mfg_record->next = tmp_mfg_record;
		  if (!prev_mfg_record)
		    first_mfg_record = mfg_record;
		  else
		    prev_mfg_record->next = mfg_record;
		  break;
		}
	      prev_mfg_record = mfg_record;
	      mfg_record = mfg_record->next;
	    }
	  if (!mfg_record)	/* last entry */
	    {
	      prev_mfg_record->next = new_mfg_record;
	      mfg_record = prev_mfg_record->next;
	    }
	}
      else
	first_mfg_record = mfg_record;
      DBG_DBG ("update_mfg_record_list: created mfg %s\n", mfg_record->name);
    }				/* if (!mfg_record) */

  /* create model entries */
  while (model)
    {
      mfg_record->model_record =
	update_model_record_list (mfg_record->model_record, model, be);
      model = model->next;
    }
  return first_mfg_record;
}




/* Create a sorted list of manufacturers based on the backends list */
static mfg_record_entry *
create_mfg_list ()
{
  mfg_record_entry *first_mfg_record = 0;
  backend_entry *be = first_backend;

  DBG_DBG ("create_mfg_list: start\n");
  while (be)
    {
      type_entry *type = be->type;
      while (type)
	{
	  /*if (type->type == dev_type)*/
	    {
	      mfg_entry *mfg = type->mfg;
	      while (mfg)
		{
		  first_mfg_record =
		    update_mfg_record_list (first_mfg_record, mfg, be);
		  mfg = mfg->next;
		}
	    }
	  type = type->next;
	}
      be = be->next;
    }
  DBG_DBG ("create_mfg_list: exit\n");
  return first_mfg_record;
}


gboolean
desc_init ()
{
  search_dir = "descriptions/";
  return read_files ();
}

GList *
desc_get_manufacturers ()
{
  mfg_record_entry *mfg_record = 0, *first_mfg_record = 0;
  GList *manufacturer_list = NULL;

  first_mfg_record = create_mfg_list ();
  mfg_record = first_mfg_record;

  while (mfg_record)
    {
      manufacturer_list = g_list_append (manufacturer_list, mfg_record->name);
      mfg_record = mfg_record->next;
    }
  return manufacturer_list;
}

GList *
desc_get_models (gchar *manufacturer_name)
{
  mfg_record_entry *mfg_record = 0, *first_mfg_record = 0;
  GList *model_list = NULL;

  first_mfg_record = create_mfg_list ();
  mfg_record = first_mfg_record;

  while (mfg_record)
    {
      if (strcmp (mfg_record->name, manufacturer_name) == 0)
	{
	  model_record_entry *model_record = mfg_record->model_record;

	  while (model_record)
	    {
	      if (model_record->status != status_unsupported)
		model_list = g_list_append (model_list, model_record);
	      model_record = model_record->next;
	    }
	  return model_list;
	}
      mfg_record = mfg_record->next;
    }
  return NULL;
}

model_record_entry *
desc_get_model (gchar * manufacturer_name, gchar *model_name)
{
  mfg_record_entry *mfg_record = 0, *first_mfg_record = 0;

  first_mfg_record = create_mfg_list ();
  mfg_record = first_mfg_record;

  while (mfg_record)
    {
      if (strcmp (mfg_record->name, manufacturer_name) == 0)
	{
	  model_record_entry *model_record = mfg_record->model_record;

	  while (model_record)
	    {
	      if (model_record->status != status_unsupported
		  && strcmp (model_record->name, model_name) == 0)
		return model_record;
	      model_record = model_record->next;
	    }
	}
      mfg_record = mfg_record->next;
    }
  return NULL;
}

